-----------------------------------------------------------------------------------------------
-- Client Lua Script for Botter
-- Copyright (c) NCsoft. All rights reserved
-- special thanks to Holicone of OneClickReport for the report implementation
-- http://www.curse.com/ws-addons/wildstar/221726-oneclickbotreport
-----------------------------------------------------------------------------------------------
require "Window"
require "GameLib"
require "Unit"
require "ChatSystemLib"
 
-----------------------------------------------------------------------------------------------
-- Botter Module Definition
-----------------------------------------------------------------------------------------------
local Botter = {} 

-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
-- hardcoded categories, hopefully Carbine does not change this
local kReportPlayerCatCount = 2
local kBotSubCatCount = 6
-- timer stuff
local kstrUpdateTimer = "BotterUpdateTimer"
local kstrPingTimer = "BotterPingTimer"
local kiPingDuration = 5
local kstrBotterVer = "Botter v1.01"
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function Botter:new(o)
	o = o or {}
	setmetatable(o, self)
	self.__index = self 

	-- initialize variables here
	o.tItems = {} -- keep track of all the list items
	o.wndSelectedListItem = nil -- keep track of which list item is currently selected
	o.tReported = {} -- keep track of reported players
	return o
end

-- init function
function Botter:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {
		-- "UnitOrPackageName",
	}
	Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

-- save function
function Botter:OnSave(eLevel)
    if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
        return nil
    end

    -- create a table to hold our data
    local tSave = {}
    for k,tItem in pairs(self.tReported) do
        -- for each player in our list, create a table using the name as our key
        tSave[k] = {}
    end
    -- simply return this value and Apollo will save the file!
    return tSave
end

-- restore function
function Botter:OnRestore(eLevel, tData)
    -- just store this and use it later
    self.tReported = tData
end
-----------------------------------------------------------------------------------------------
-- Botter OnLoad
-----------------------------------------------------------------------------------------------
function Botter:OnLoad()
	-- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("Botter.xml")
	self.wndMain = Apollo.LoadForm(self.xmlDoc, "BotterForm", nil, self)
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

-----------------------------------------------------------------------------------------------
-- Botter OnDocLoaded
-----------------------------------------------------------------------------------------------
function Botter:OnDocLoaded()
	-- make sure doc is loaded before doing anything!
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then        
		if self.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end

		self.wndReportForm = Apollo.LoadForm(self.xmlDoc, "ReportForm", nil, self)
		--make sure sub-doc is loaded before doing anything!
		if self.wndReportForm == nil then
			Apollo.AddAddonErrorText(self, "Could not load the child window for some reason.")
			return
		end

		-----------------------------------------------------------------------------------
		-- main window
		self.wndItemList = self.wndMain:FindChild("ItemList")
		self.wndMyPos = self.wndMain:FindChild("MyLocation")
		self.wndRealTime = self.wndMain:FindChild("RealTime")
		self.wndRealTime:SetBGColor("blue")
		self.wndOnOff = self.wndMain:FindChild("OnOff")
		self.wndOnOff:SetBGColor("red")
		self.strZoneName = "Unknown"
		self.unknown = true
		self.scrollList = true

		-----------------------------------------------------------------------------------
		-- child window
		self.wndEdit = self.wndReportForm:FindChild("etReport")
		self.wndIndicator = Apollo.LoadForm(self.xmlDoc, "Indicator", nil, self)
		self.wndTitle = self.wndReportForm:FindChild("Title")

		-----------------------------------------------------------------------------------
		-- visibility
		self.focus = false
		self.update = false
		self.realtime = false
		self.wndMain:Show(false, true)
		self.wndReportForm:Show(false, true)
		self.wndIndicator:Show(false)
		self.wndMain:SetStyle("AutoFadeBG", true)
		self.wndMain:SetStyle("AutoFadeNC", true)

		-----------------------------------------------------------------------------------
		-- Register handlers for events, slash commands and timer, etc.
		Apollo.RegisterSlashCommand("botter", "OnBotterOn", self)

		--zone change handler
		Apollo.RegisterEventHandler("VarChange_ZoneName", "OnChangeZoneName", self)

		--timer handlers
		Apollo.RegisterTimerHandler(kstrUpdateTimer, "OnUpdateTimer", self)
		self.timer = ApolloTimer.Create(1.0, true, "OnUpdateTimer", self)
		Apollo.RegisterTimerHandler(kstrPingTimer, "OnPingTimer", self)
		Apollo.CreateTimer(kstrPingTimer, kiPingDuration, false)
		Apollo.StopTimer(kstrPingTimer)
	end
end

-----------------------------------------------------------------------------------------------
-- Botter Functions
-----------------------------------------------------------------------------------------------
-- on SlashCommand "/botter" or "/botter (argument)"
function Botter:OnBotterOn(cmd, args)
	local strMessage = ""

	-- check for arguments, change appropriately, and set appropriate text
	if args == "on" then
		self:EnableTracking()
		strMessage = "Botter On"
	elseif args == "off" then
		self:DisableTracking()
		strMessage = "Botter Off"
	elseif args == "log" then
		self:DisableRealTime()
		strMessage = "Botter mode: Log"
	elseif args == "realtime" then
		self:EnableRealTime()
		strMessage = "Botter mode: RealTime"
	else -- no argument used, don't send message
		self.wndMain:Invoke() -- show the window
		return
	end

	-- an argument was used, so notify user of state change
	ChatSystemLib.PostOnChannel( ChatSystemLib.ChatChannel_Command, strMessage, "" )
end

-- When the player changes zones, keep track of the current zone they are in (for list)
function Botter:UpdateItemZone()
	-- we don't know where we are yet, never mind
	if self.strZoneName == nil or self.strZoneName == "Unknown" then
		return
	end

	-- otherwise, try to fix name list
	for idx, strZoneName in pairs(self.tItems) do
		if strZoneName ~= nil and strZoneName == "Unknown" then
			strZoneName = self.strZoneName
			local i = self.tItems[idx]
	   		if i.wnd ~= nil then
        		self:UpdateItem(i)
			end
		end
	end

	-- we know where we are
	self.unknown = false
end

-- Figure out what zone the player is in right now
function Botter:GetCurrentZoneName()
	local currentZone = GameLib.GetCurrentZoneMap()
	if currentZone == nil then
		self.unknown = true
		return "Unknown"
	else
		return currentZone.strName
	end
end

-- determine color
function Botter:GetColor(unit)
	--default color
	if self.tReported[unit:GetName()] ~= nil then
		return "red"
	end

	if self.unknown then
		return "gray"
	end
	local myFaction = GameLib.GetPlayerUnit():GetFaction() == unit:GetFaction()
	local myGuild = GameLib.GetPlayerUnit():GetAffiliationName() == unit:GetAffiliationName()
	local pvp = unit:IsPvpFlagged()
	local myFriend = unit:IsFriend() or unit:IsAccountFriend()

	--differentiate friends
	if myFriend then
		return "blue"
	end

	if myGuild then
		return "ff913db7"
	end

	--other faction or mine?
	if myFaction then
		if pvp then
			return "green"
		end
		return "white"
	else
		if pvp then
			return "ffd18046"
		end
		return "yellow"
	end

	return "gray"
end
-----------------------------------------------------------------------------------------------
-- Message Handlers
-----------------------------------------------------------------------------------------------
--handles unit created: a unit has come in range of the player
function Botter:OnUnitCreated(unit)
	-- make sure the unit created is a character, or this list will
	-- get filled up with a lot of stuff
	if not unit:IsACharacter() then
		return
	end

	-- add unit to list and fill in some information about this unit
	local strName = unit:GetName()
	if self.tItems[strName] == nil then
		self.tItems[strName] = {}
		self.tItems[strName].strName = strName
	end

	-- fill the table for the unit
	local i = self.tItems[strName]
	i.nLevel = unit:GetLevel()
	i.nId = unit:GetId()
	local pos = unit:GetPosition()
	i.pos = pos
	i.strLocation = string.format("%d, %d, %d", pos.x, pos.z, pos.y)
	i.strZoneName = self:GetCurrentZoneName()
	i.cColor = self:GetColor(unit)

	-- do we already know about this unit? If so, just update it
	if i.wnd ~= nil then
		self:UpdateItem(i)
		return
	end

	-- assign gray to be the color of this new unit
	--i.crDot = ApolloColor.new("gray")

	-- store the info about this player using the name as the key
	self.tItems[i.strName] = i

	-- add this item to our Window
	self:AddItem(i)
end

--handles unit destroyed: a unit has left range of the player
function Botter:OnUnitDestroyed(unit)
	-- make sure the unit created is a character, or this list will
	-- get filled up with a lot of stuff
	if not unit:IsACharacter() then
		return
	end
	-- remove unit from list only if that player is on it already
	local strName = unit:GetName()
	if self.tItems[strName] ~= nil then
		local i = self.tItems[strName]
		self.tItems[strName] = nil
		if i.wnd ~= nil then
			self:RemoveItem(i)
			self.wndItemList:ArrangeChildrenVert()
		end
	end
end

-- called when the player changes zone, sets zone name for new members
function Botter:OnChangeZoneName(oVar, strNewZone)
	if strNewZone == nil then
		return
	end
	--self.strZoneName = strNewZone
	self.strZoneName = self:GetCurrentZoneName()
end

-----------------------------------------------------------------------------------------------
-- Timer Handlers
-----------------------------------------------------------------------------------------------
-- on update timer; currently it is not disabled but continues to run
function Botter:OnUpdateTimer()
	-- see if we're actually updating
	if not self.wndMain:IsVisible() then
		return
	end

	-- figure out where we are
	self.strZoneName = self:GetCurrentZoneName()

	-- update player position if possible
	if GameLib.GetPlayerUnit() ~= nil then	
		local pos = GameLib.GetPlayerUnit():GetPosition()
		local t = string.format("%s: %d, %d, %d", self.strZoneName, pos.x, pos.z, pos.y)
		self.wndMyPos:SetText(t)
	end
	
	-- attempt to correct issue where loading zone while on will not show  correct zone
	if self.unknown then
		self:UpdateItemZone()
	end
end

-- removes visible marker after a time
function Botter:OnPingTimer()
	Apollo.StopTimer(kstrPingTimer)
	self.wndIndicator:Show(false)
end

-----------------------------------------------------------------------------------------------
-- ItemList Functions
-----------------------------------------------------------------------------------------------
-- add a player into the item list
function Botter:AddItem(i)
	-- load the window item for the list item
	local wnd = Apollo.LoadForm(self.xmlDoc, "ListItem", self.wndItemList, self)
	-- update the item
	i.wnd = wnd
	self:UpdateItem(i)
	self.wndItemList:ArrangeChildrenVert()
	if self.scrollList then
		self.wndItemList:SetVScrollPos(self.wndItemList:GetVScrollRange())
	end
end

-- update a player already in the item list
function Botter:UpdateItem(i)
	-- fill in our data
	i.wnd:FindChild("PlayerName"):SetTextColor(i.cColor)
	i.wnd:FindChild("PlayerName"):SetText(i.strName)
	i.wnd:FindChild("PlayerLevel"):SetText(i.nLevel)
	i.wnd:FindChild("PlayerZone"):SetText(i.strZoneName)
	i.wnd:FindChild("PlayerLocation"):SetText(i.strLocation)
end

-- remove a player from the list
function Botter:RemoveItem(i)
	i.wnd:Destroy()
end

---------------------------------------------------------------------------------------------------
-- Botter Form Functions
---------------------------------------------------------------------------------------------------
-- when the Cancel button is clicked (the red button in upper right)
function Botter:OnCancel()
	self.wndMain:Close() -- hide the window
end

--Change mode to RealTime, which removes players from the list who go away
function Botter:EnableRealTime( wndHandler, wndControl, eMouseButton )
	Apollo.RegisterEventHandler("UnitDestroyed", "OnUnitDestroyed", self)
	self.wndRealTime:SetBGColor("blue")
	self.realtime = true
	if self.focus then
		self.wndRealTime:SetText("RealTime")
	end
end

--Change mode to Log, which adds any player to the list who approach
function Botter:DisableRealTime( wndHandler, wndControl, eMouseButton )
	Apollo.RemoveEventHandler("UnitDestroyed", self)
	self.wndRealTime:SetBGColor("yellow")
	self.realtime = false
	if self.focus then
		self.wndRealTime:SetText("Log")
	end
end

--Turn off listening for new units
function Botter:DisableTracking( wndHandler, wndControl, eMouseButton )
	Apollo.RemoveEventHandler("UnitCreated", self)
	Apollo.RemoveEventHandler("UnitUpdated", self)
	self.update = false
	self.wndOnOff:SetBGColor("red")
	if self.focus then
		self.wndOnOff:SetText("Off")
	end
end

--Turn on listening for new units
function Botter:EnableTracking( wndHandler, wndControl, eMouseButton )
	Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", self)
	Apollo.RegisterEventHandler("UnitUpdated", "OnUnitCreated", self)
	self.update = true
	self.wndOnOff:SetBGColor("green")
	if self.focus then
		self.wndOnOff:SetText("On")
	end
end

---------------------------------------------------------------------------------------------------
-- ListItem Functions
---------------------------------------------------------------------------------------------------
-- this is the REPORT button
function Botter:OnGenerateReport( wndHandler, wndControl, eMouseButton )
	-- make sure the wndControl is valid
	if wndHandler ~= wndControl then
		return
	end

	-- get our suspect
	local target = wndControl:GetParent():FindChild("PlayerName"):GetText()

	-- suspect does not exist, shouldn't ever hit this
	if self.tItems[target] == nil then
		return
	end

	-- we found the player, setup form
	self.wndEdit:SetText(self:GetReportText(self.tItems[target]))
	self:SetTitle(target)
	self.wndReportForm:Invoke() --Show(true, true)

	--debugging
	--Apollo.GetAddon("ChatLog"):HelperGetCurrentEditbox():InsertText(wndControl:GetParent():FindChild("PlayerLocation"):GetText())
end

-- this is the LOCATE button
function Botter:OnLocatePlayer( wndHandler, wndControl, eMouseButton )
	local target = wndControl:GetParent():FindChild("PlayerName"):GetText()
	self.wndIndicator:Show(false)

	-- cant find the list member, shouldn't ever hit this
	if self.tItems[target] == nil then
		return
	end

	-- can't find the list member in the world
	local unit = GameLib.GetUnitById(self.tItems[target].nId)
	if unit == nil then
		self:NotifyDirection(self.tItems[target])
		return
	end

	-- add minimap ping
	Event_FireGenericEvent("ZoneMapPing", unit:GetId(), unit:GetPosition())

	-- we found the player, this part is sketchy - try to show arrow
	if not unit:IsOccluded() then
		self.wndIndicator:SetUnit(unit, 1)
		self.wndIndicator:Show(true)

		-- set timer to remove indicator after a few seconds
		Apollo.StartTimer(kstrPingTimer)
	else
		unit:ShowHintArrow()
		--self:NotifyDirection(unit)
	end
end

---------------------------------------------------------------------------------------------------
-- ReportForm Functions
---------------------------------------------------------------------------------------------------
-- Cancel the report and clear generated/added text
function Botter:OnCancelReport( wndHandler, wndControl, eMouseButton )
	self.wndReportForm:Close() --Show(false, true)
	--make sure we can modify the box
	if self.wndEdit == nil then
		Apollo.AddAddonErrorText(self, "Could not load report box.")
		return
	end

	--clear the text
	self.wndEdit:SetText("")
end

-- Submit the report if it is not empty
function Botter:OnReport( wndHandler, wndControl, eMouseButton )
 	--make sure we can read the box
	if self.wndEdit == nil then
		Apollo.AddAddonErrorText(self, "Could not load report box.")
		return
	end

	--make sure the text isn't empty
	if self.wndEdit:GetText() ~= "" then
		PlayerTicketDialog_Report(kReportPlayerCatCount, kBotSubCatCount, self.wndEdit:GetText())
	end

	-- add player to list of reported
	if self.target ~= nil then
		self.tReported[self.target] = {}
	end

	--clear the text
	self.wndEdit:SetText("")
	self.wndReportForm:Close() --Show(false, true)
end

-- Generate text for report based on things Botter knows
function Botter:GetReportText(unit)
	local reportText = "Botter AddOn Generated Report\nPlayer: " .. unit.strName
	reportText = reportText .. "\nLevel: " .. unit.nLevel
	reportText = reportText .. "\nZone: " .. unit.strZoneName
	reportText = reportText .. "\nX,Z,Y: " .. unit.strLocation
	reportText = reportText .. "\nAdditional information:\n"
	return(reportText)
end

-- set the title based on if we've reported this player before
function Botter:SetTitle(strPlayer)
	local title = ""
	if self.tReported[strPlayer] == nil then
		title = "Report " .. strPlayer .. "?"
		self.target = strPlayer
	else
		title = "You have already reported " .. strPlayer .. "!"
		self.target = nil
	end
	self.wndTitle:SetText(title)
end

-- notify player where target is
function Botter:NotifyDirection(item)
	local strMessage = "Botter: " .. item.strName .. " is "
	local unitPos = item.pos
	local myPos = GameLib.GetPlayerUnit():GetPosition()
	local rawZ = myPos.z - unitPos.z
	if rawZ < 0 then
		rawZ = rawZ * -1
	end
	local rawX = myPos.x - unitPos.x
	if rawX < 0 then
		rawX = rawX * -1
	end
	if rawX < (rawZ * 2) then
		if myPos.z < unitPos.z then
			strMessage = strMessage .. "south"
		else
			strMessage = strMessage .. "north"
		end
	end
	if rawZ < (rawX * 2) then
		if myPos.x < unitPos.x then
			strMessage = strMessage .. "east"
		else
			strMessage = strMessage .. "west"
		end
	end
	ChatSystemLib.PostOnChannel( ChatSystemLib.ChatChannel_Command, strMessage, "" )
end

---------------------------------------------------------------------------------------------------
-- BotterForm Functions
---------------------------------------------------------------------------------------------------

function Botter:OnBtnOnOff()
	if "On" == self.wndOnOff:GetText() then
		self:DisableTracking()
	else
		self:EnableTracking()
	end
end

function Botter:OnRealTime()
	if "RealTime" == self.wndRealTime:GetText() then
		self:DisableRealTime()
	else
		self:EnableRealTime()
	end
end

function Botter:OnMouseEnter( wndHandler, wndControl, x, y )
	if wndHandler:GetId() ~= wndControl:GetId() then
		return
	end
	self.scrollList = false
	local bContainsMouse = wndHandler:GetParent():ContainsMouse()
	if wndHandler:GetData() and not bContainsMouse then -- Because UseParentOpacity is BGColor not TextColor
		for idx, wndCurr in pairs(wndHandler:GetData()) do
			wndCurr:Show(false)
		end
	end

end

function Botter:OnMouseExit( wndHandler, wndControl, x, y )
	if wndHandler:GetId() ~= wndControl:GetId() then
		return
	end
	self.scrollList = true
end

function Botter:OnMouseEnterMain( wndHandler, wndControl, x, y )
	local temp = "Off"
	if self.update then
		temp = "On"
	end
	self.wndOnOff:SetText(temp)
	temp = "Log"
	if self.realtime then
		temp = "RealTime"
	end
	self.wndRealTime:SetText(temp)
	local wnd = self.wndMain:FindChild("Title")
	wnd:SetText(kstrBotterVer)
	wnd = self.wndMain:FindChild("Text")
	wnd:SetText("Toggles")
	self.focus = true
end

function Botter:OnMouseExitMain( wndHandler, wndControl, x, y )
	self.focus = false
	self.wndOnOff:SetText("")
	self.wndRealTime:SetText("")
	local wnd = self.wndMain:FindChild("Title")
	wnd:SetText("")
	wnd = self.wndMain:FindChild("Text")
	wnd:SetText("")
end

function Botter:OnClear( wndHandler, wndControl, eMouseButton )
	self.wndItemList:DestroyChildren()
	self.wndSelectedListItem = nil 
	self.tItems = {}
	self.wndItemList:ArrangeChildrenVert()
	self.wndIndicator:Show(false)
end

-----------------------------------------------------------------------------------------------
-- Botter Instance
-----------------------------------------------------------------------------------------------
local BotterInst = Botter:new()
BotterInst:Init()
